"""Delta F over F calculations."""

import numpy as np
import pandas as pd


def compute_delta_f_over_f(
        movie: np.ndarray,
        baselines: pd.DataFrame,
        start_offset: int = 0,
        end_offset: int = 0,
        method: str = "mean",
        apply_limits: bool = True,
        absolute_limit: float = 10.0,
) -> np.ndarray:
    """Compute delta F over F movies based on baselines."""
    n_frames = movie.shape[0]
    n_baselines = baselines.shape[0]
    movie = movie.astype(np.float32)
    for i_row, row in baselines.iterrows():
        baseline_start = row["first_frame"] + start_offset
        baseline_end = row["last_frame"] + 1 + end_offset

        current_baseline = row["i_baseline"]
        next_baseline = current_baseline + 1
        is_next = baselines["i_baseline"] == next_baseline

        if current_baseline == 0:
            print("First baseline scales movie from beginning.")
            scale_start = 0
        else:
            scale_start = baseline_start
        if current_baseline == (n_baselines - 1):
            print("Last baseline scales movie until ending.")
            scale_end = n_frames
        else:
            scale_end = baselines.loc[is_next, "first_frame"].values[0]

        n_baseline = baseline_end - baseline_start
        n_scale = scale_end - scale_start

        baseline_movie = movie[baseline_start:baseline_end, :, :]
        scale_movie = movie[scale_start:scale_end]

        if method == "mean":
            baseline_image = np.mean(baseline_movie, axis=0)
        elif method == "median":
            baseline_image = np.median(baseline_movie, axis=0)
        else:
            raise ValueError(f"{method=} unknown")
        scale_movie = (scale_movie - baseline_image) / baseline_image

        if apply_limits:
            scale_movie[scale_movie < -absolute_limit] = -absolute_limit
            scale_movie[scale_movie > absolute_limit] = absolute_limit
        print(f"Baseline {current_baseline} {baseline_start}:{baseline_end} ({n_baseline}) scales {scale_start}:{scale_end} ({n_scale})")
        print(f"\tAfter scaling: from {np.nanmin(scale_movie):.1f} to {np.nanmax(scale_movie):.1f} dF/F ({apply_limits=}, {absolute_limit=:.1f})")
        movie[scale_start:scale_end, :, :] = scale_movie
    return movie



"""Stimulus triggered averages."""

import numpy as np
import pandas as pd


def compute_stimulus_triggered_average(movie: np.ndarray, stim_periods: pd.DataFrame, start_offset: int = 0, end_offset: int = 0) -> np.ndarray:
    """Compute stimulus triggered average from stim periods."""
    repeat_movies = []
    for i_row, row in stim_periods.iterrows():
        start_index = row["first_frame"] + start_offset
        end_index = row["last_frame"] + 1 + end_offset
        single_repeat = movie[start_index:end_index, :, :]
        repeat_movies.append(single_repeat)
    min_length = np.min([single_movie.shape[0] for single_movie in repeat_movies])
    repeat_movies = [single_movie[:min_length, :, :] for single_movie in repeat_movies]
    repeat_tensor = np.stack(repeat_movies, axis=0)
    n_repeats = repeat_tensor.shape[0]
    print(f"Averaging over {n_repeats} repeats of {min_length} frames.")
    sta = np.mean(repeat_tensor, axis=0)
    return sta



"""Resample movies e.g. by downsizing."""

import numpy as np
from skimage.transform import downscale_local_mean
from tqdm import tqdm


def downsize_movie_with_local_mean(movie: np.ndarray, factor: int) -> np.ndarray:
    """Downsize each frame using skimage.transform.downscale_local_mean."""
    n_frames, x, y = movie.shape
    new_x = int(np.rint(x / factor))
    new_y = int(np.rint(y / factor))
    downsized_movie = np.zeros((n_frames, new_x, new_y))
    print("Downsizing movie.")
    for i_frame in tqdm(range(n_frames)):
        original_frame = movie[i_frame, :, :]
        smaller_frame = downscale_local_mean(original_frame, (factor, factor))
        downsized_movie[i_frame, :, :] = smaller_frame
    return downsized_movie



"""Interact with local file elements."""

from pathlib import Path

import numpy as np
import tifffile
from tqdm import tqdm


def list_files_and_folders(local_folder: Path) -> tuple:
    """Get contents of a folder sorted into files and folders."""
    contents = local_folder.iterdir()
    files = []
    folders = []
    for element in contents:
        if element.is_file():
            files.append(element)
        elif element.is_dir():
            folders.append(element)
        else:
            raise ValueError(f"{element} is neither file nor folder?")
    return files, folders


def identify_tifs(files: list[Path]) -> list:
    """Identify files that end with .tif in a list of Path objects."""
    tif_files = []
    for file in files:
        if file.suffix in [".tif", ".tiff"]:
            tif_files.append(file)
    return tif_files


def convert_tif_folder_into_file(source_folder: Path,  tif_type: str, output_file: Path | None = None) -> Path:
    """Convert a folder full of single frame tifs into a single multi-frame tif file."""
    if not source_folder.is_dir():
        raise FileNotFoundError(f"{source_folder} does not exist.")

    if output_file is None:
        target_parent = source_folder.parent
        output_file = target_parent / f"{source_folder.name}.tif"
        print(f"No output file given - using {output_file}")
    else:
        assert output_file.suffix == ".tif"
    if output_file.is_file():
        raise FileExistsError(f"{output_file} already exists.")

    tifs = list_tifs(source_folder)
    print(f"{len(tifs)} found in {source_folder}.")
    sorted_tifs = sort_tifs(tifs, tif_type)
    print(f"Loading {sorted_tifs.size} tifs.")
    movie = load_tifs(sorted_tifs)
    print(f"Tifs loaded into movie with shape {movie.shape}.")
    print(f"Saving movie to {output_file}.")
    write_movie_to_tif(movie, output_file)
    return output_file


def list_tifs(local_folder: Path) -> list:
    """List all .tif files in a folder."""
    files, _ = list_files_and_folders(local_folder)
    tif_files = identify_tifs(files)
    return tif_files


def write_movie_to_tif(movie: np.ndarray, target_file: Path) -> None:
    """Write a 3D numpy array (frames, x, y) to a .tif file."""
    tifffile.imwrite(target_file, data=movie, bigtiff=True)


def load_tifs(tif_files: list[Path], dtype=np.uint8) -> np.ndarray:
    """Load single frame tifs into a single 3D numpy array (frames, x, y)."""
    list_of_images = []
    for file in tqdm(tif_files):
        image = tifffile.imread(file).astype(dtype)
        list_of_images.append(image)
    movie = np.stack(list_of_images, axis=0)
    return movie


def sort_tifs(tif_files: list[Path], tif_type: str) -> np.ndarray[Path]:
    """Sort tifs by file name."""
    match tif_type:
        case "leica":
            sorted_tifs = sort_leica_tifs(tif_files)
        case "basler":
            sorted_tifs = sort_basler_tifs(tif_files)
        case _:
            raise ValueError(f"{tif_type} unknown.")
    return sorted_tifs


def sort_leica_tifs(tif_files: list[Path]) -> np.ndarray[Path]:
    """
    Example file name: mapping_with_shielding_t0000_RAW_ch00.tif
    """
    file_names = [file.name for file in tif_files]
    sort_vector = np.argsort(file_names)
    sorted_files = np.asarray(tif_files)[sort_vector]
    return sorted_files


def sort_basler_tifs(tif_files: list[Path]) -> np.ndarray[Path]:
    """
    Example file name: Basler_a2A1920-160umPRO__40349470__20230509_141623734_0000.tiff
    Unfortunately, numbers are not fully zero-padded so sorting needs to be changed a little.
    """
    file_stems = [file.stem for file in tif_files]
    stem_endings = [stem.split("_")[-1] for stem in file_stems]
    stem_endings = [int(ending) for ending in stem_endings]
    sort_vector = np.argsort(stem_endings)
    sorted_files = np.asarray(tif_files)[sort_vector]
    return sorted_files



"""Bandpass filter widefield movies."""

import numpy as np
from scipy import signal


def apply_bandpass_filter(movie: np.ndarray, frame_rate: float, start_freq: float, end_freq: float, filter_order: int = 1) -> np.ndarray:
    """Apply bandpass filter to movie."""
    butterworth_filter = signal.butter(
        N=filter_order,
        Wn=[start_freq, end_freq],
        output="sos",
        btype="bandpass",
        fs=frame_rate,
    )
    filtered_movie = signal.sosfiltfilt(butterworth_filter, movie, axis=0)
    return filtered_movie


def remove_high_frequencies(movie: np.ndarray, frame_rate: float, cutoff_freq: float, filter_order: int = 1) -> np.ndarray:
    """Remove high frequencies from movie."""
    butterworth_filter = signal.butter(
        N=filter_order,
        Wn=cutoff_freq,
        output="sos",
        btype="lowpass",
        fs=frame_rate,
    )
    filtered_movie = signal.sosfilt(butterworth_filter, movie, axis=0)
    return filtered_movie


def remove_low_frequencies(movie: np.ndarray, frame_rate: float, cutoff_freq: float, filter_order: int = 1) -> np.ndarray:
    """Remove low frequencies from movie."""
    butterworth_filter = signal.butter(
        N=filter_order,
        Wn=cutoff_freq,
        output="sos",
        btype="highpass",
        fs=frame_rate,
    )
    filtered_movie = signal.sosfilt(butterworth_filter, movie, axis=0)
    return filtered_movie



"""Smooth movies with gaussian kernels."""

import numpy as np
from scipy.ndimage import gaussian_filter
from tqdm import tqdm


def smooth_spatially(raw_movie: np.ndarray, sigma: float) -> np.ndarray:
    """Apply gaussian kernel convolution to each frame separately."""
    smoothed_movie = np.zeros_like(raw_movie)
    n_frames = raw_movie.shape[0]
    for i_frame in tqdm(range(n_frames)):
        smoothed_frame = gaussian_filter(raw_movie[i_frame, :, :], sigma=sigma)
        smoothed_movie[i_frame, :, :] = smoothed_frame
    return smoothed_movie


def smooth_temporally(movie: np.ndarray, sigma: float) -> np.ndarray:
    raise NotImplementedError()


def smooth_all_dimensions(movie: np.ndarray, sigma: float) -> np.ndarray:
    """Apply gaussian kernel convolution along all dimensions of a movie."""
    movie = gaussian_filter(movie, sigma=sigma)
    return movie